from functools import partial

import numpy as np
import pyopencl as cl
import pyopencl.array as cl_array
from pyopencl import Image
from pyopencl.array import Array as ClArray
from pyvkfft.fft import irfftn, rfftn
from scipy.ndimage import laplace as laplace_filter
from tqdm import tqdm

from powerfit_em.correlators.clkernels import CLKernels
from powerfit_em.correlators.shared import Correlator, Vars, VarsFT, f32, get_ft_shape, get_lcc_mask, i32


def init_gpu_vars(
    queue: cl.CommandQueue,
    target: np.ndarray,
    laplace: bool,
) -> tuple[Vars[ClArray, Image], VarsFT[ClArray]]:
    """Initialize all GPU variables on the specified queue."""
    lcc_mask = get_lcc_mask(target)
    _t = laplace_filter(target, mode="wrap") if laplace else target
    zeros = np.zeros(target.shape, f32)
    gpu_vars = Vars(
        target=cl_array.to_device(queue, _t.astype(f32)),
        template=cl.image_from_array(queue.context, zeros),  # template is set through separate method
        mask=cl.image_from_array(queue.context, zeros),  # mask is set through separate method
        lcc_mask=cl_array.to_device(queue, lcc_mask.astype(i32)),
        target2=cl_array.to_device(queue, zeros),
        rot_template=cl_array.to_device(queue, zeros),
        rot_mask=cl_array.to_device(queue, zeros),
        rot_mask2=cl_array.to_device(queue, zeros),
        gcc=cl_array.to_device(queue, zeros),
        ave=cl_array.to_device(queue, zeros),
        ave2=cl_array.to_device(queue, zeros),
        lcc=cl_array.to_device(queue, zeros),
        rot=cl_array.to_device(queue, np.zeros(target.shape, i32)),
    )
    zeros_ft = np.zeros(get_ft_shape(target), dtype=np.complex64)
    gpu_vars_ft = VarsFT(
        target=cl_array.to_device(queue, zeros_ft),
        target2=cl_array.to_device(queue, zeros_ft),
        template=cl_array.to_device(queue, zeros_ft),
        mask=cl_array.to_device(queue, zeros_ft),
        mask2=cl_array.to_device(queue, zeros_ft),
        ave=cl_array.to_device(queue, zeros_ft),
        ave2=cl_array.to_device(queue, zeros_ft),
        lcc=cl_array.to_device(queue, zeros_ft),
        gcc=cl_array.to_device(queue, zeros_ft),
    )
    return gpu_vars, gpu_vars_ft


def generate_kernels(queue: cl.CommandQueue, target: np.ndarray):
    """Generate the custom OpenCL kernels based on the target's shape"""
    kernel_values = {
        "shape_x": target.shape[2],
        "shape_y": target.shape[1],
        "shape_z": target.shape[0],
        "llength": i32(min(target.shape) // 2),
    }
    return CLKernels(queue.context, kernel_values)


def precompute_squared_targets(gpu_vars: Vars[ClArray, Image], gpu_vars_ft: VarsFT[ClArray], kernels: CLKernels):
    """Compute the squared target and fourier transformed target on GPU for reuse."""
    gpu_vars_ft.target = rfftn(gpu_vars.target)
    kernels.multiply(gpu_vars.target, gpu_vars.target, gpu_vars.target2)
    rfftn(gpu_vars.target2, gpu_vars_ft.target2)


def transform_rotations(rotations: np.ndarray) -> np.ndarray:
    """Transform rotation array for input into OpenCL kernels.

    The OpenCL kernel requires a Float16 input (struct containing 16 single-
    precision floats). The rotation matrices need to occupy the first 9 entries.
    """
    rot_trans = np.zeros((rotations.shape[0], 16), dtype=np.float32)
    rot_trans[:, :9] = rotations.reshape(-1, 9)
    return rot_trans


class GPUCorrelator(Correlator):
    """Compute the LCC score for a target and template combination."""

    def __init__(
        self,
        target: np.ndarray,
        template: np.ndarray,
        rotations: np.ndarray,
        mask: np.ndarray,
        queue: cl.CommandQueue,
        laplace: bool = False,
    ):
        """Initialize the GPU correlator.

        Args:
            target: the target density on which you want to fit a template structure.
            template: the template structure that you want to fit in the target density,
                should have been regridded to the same grid as the target density.
            rotations: array of 3D-rotation matrices, of shape (n_rotations, 3, 3).
                for each rotation the local cross correlation is computed for every
                possible translation (in parallel using Fourier transforms).
            mask: core-weighted mask. See doi:10.3934/biophy.2015.2.73, Figure 1.
            queue: the OpenCL command queue on which to execute the computations.
            laplace: if true, a Laplace pre-filter is applied to the target density and
                template to enhance the sensitivity of the scoring function.
        """
        self.target: np.ndarray = target / target.max()
        self.laplace = laplace
        self.mask = mask
        self.queue = queue
        self.norm_factor = 0.0  # to be set by set_template

        self._rotations = transform_rotations(rotations)

        self.vars, self.vars_ft = init_gpu_vars(queue, self.target, self.laplace)

        self.lcc = np.zeros(self.target.shape, dtype=np.float32)
        self.rot = np.zeros(self.target.shape, dtype=np.int32)

        self.set_template(template, mask)

        self.cl_kernels = generate_kernels(queue, self.target)
        self.conj_multiply = self.cl_kernels.conj_multiply
        self.square = lambda a, b: self.cl_kernels.multiply(a, a, b)
        self.rfftn = rfftn
        self.irfftn = irfftn
        precompute_squared_targets(self.vars, self.vars_ft, self.cl_kernels)

    def _set_template_var(self, template: np.ndarray):
        self.vars.template = cl.image_from_array(self.vars.template.context, template.astype(f32))

    def _set_mask_var(self, mask: np.ndarray):
        self.vars.mask = cl.image_from_array(self.vars.mask.context, mask.astype(f32))

    def rotate_grids(self, rotmat: np.ndarray):
        """Rotate the template and mask using the rotational matrix."""
        self.cl_kernels.rotate_image3d(self.queue, self.vars.template, rotmat, self.vars.rot_template)
        self.cl_kernels.rotate_image3d(self.queue, self.vars.mask, rotmat, self.vars.rot_mask, nearest=True)

    def compute_lcc_score_and_take_best(self, n: int):
        """Compute the LCC score and store best result.

        Args:
            n: iteration number.
        """
        self.cl_kernels.calc_lcc_and_take_best(
            self.vars.gcc,
            self.vars.ave,
            self.vars.ave2,
            self.vars.lcc_mask,
            self.norm_factor,
            np.int32(n),
            self.vars.lcc,
            self.vars.rot,
        )

    def retrieve_results(self):
        """Retrieve the results from the GPU."""
        self.vars.lcc.get(ary=self.lcc)
        self.vars.rot.get(ary=self.rot)
        self.queue.finish()

    def scan(self, progress: partial[tqdm] | None):
        """Scan all provided rotations to find the best fit."""
        self.vars.lcc.fill(0)
        self.vars.rot.fill(0)

        _range = range(0, self._rotations.shape[0])
        if progress is None:
            for n in _range:
                self.compute_rotation(n, self._rotations[n])
        else:
            for n in progress(_range):
                self.compute_rotation(n, self._rotations[n])
                self.queue.finish()
        self.retrieve_results()



import unittest

import numpy as np

from powerfit_em.helpers import opencl_available

if opencl_available():
    import pyopencl as cl
    import pyopencl.array as cl_array

    from powerfit_em.correlators.clkernels import CLKernels


@unittest.skipIf(not opencl_available(), "GPU resources are not available.")
class TestCLKernels(unittest.TestCase):
    """Tests for the OpenCL kernels"""

    def setUp(self):
        p = cl.get_platforms()[0]
        devs = p.get_devices()
        self.ctx = cl.Context(devices=devs)
        self.queue = cl.CommandQueue(self.ctx, device=devs[0])
        values = {
            "shape_x": 10,
            "shape_y": 0,
            "shape_z": 0,
            "llength": 5,
        }
        self.k = CLKernels(self.ctx, values=values)
        self.s_linear = cl.Sampler(self.ctx, False, cl.addressing_mode.CLAMP, cl.filter_mode.LINEAR)
        self.s_nearest = cl.Sampler(self.ctx, False, cl.addressing_mode.CLAMP, cl.filter_mode.NEAREST)

    def test_multiply(self):
        np_in1 = np.arange(10, dtype=np.float32)
        np_in2 = np.arange(10, dtype=np.float32)
        np_out = np_in1 * np_in2

        cl_in1 = cl_array.to_device(self.queue, np_in1)
        cl_out = cl_array.to_device(self.queue, np.zeros(10, dtype=np.float32))
        cl_in2 = cl_array.to_device(self.queue, np_in2)

        self.k.multiply(cl_in1, cl_in2, cl_out)
        self.assertTrue(np.allclose(np_out, cl_out.get()))

    def test_conj_multiply(self):
        np_in1 = np.zeros(10, dtype=np.complex64)
        np_in2 = np.zeros(10, dtype=np.complex64)
        np_in1.real = np.random.rand(10)
        np_in1.imag = np.random.rand(10)
        np_in2.real = np.random.rand(10)
        np_in2.imag = np.random.rand(10)
        np_out = np_in1.conj() * np_in2

        cl_in1 = cl_array.to_device(self.queue, np_in1)
        cl_in2 = cl_array.to_device(self.queue, np_in2)
        cl_out = cl_array.to_device(self.queue, np.zeros(10, dtype=np.complex64))
        self.k.conj_multiply(cl_in1, cl_in2, cl_out)
        self.assertTrue(np.allclose(np_out, cl_out.get()))


if __name__ == "__main__":
    unittest.main()



import gzip

import pytest

from powerfit_em.structure import parse_mmcif, parse_pdb


@pytest.fixture
def sample_cif():
    return """\
data_6J5VA2A    
loop_
_atom_site.group_PDB
_atom_site.id
_atom_site.type_symbol
_atom_site.label_atom_id
_atom_site.label_alt_id
_atom_site.label_comp_id
_atom_site.label_asym_id
_atom_site.label_entity_id
_atom_site.label_seq_id
_atom_site.pdbx_PDB_ins_code
_atom_site.Cartn_x
_atom_site.Cartn_y
_atom_site.Cartn_z
_atom_site.occupancy
_atom_site.B_iso_or_equiv
_atom_site.pdbx_formal_charge
_atom_site.auth_seq_id
_atom_site.auth_asym_id
_atom_site.pdbx_PDB_model_num
ATOM 1 N N . MET A 1 1 ? 88.638 99.772 125.454 1 165.94 ? 1 A 1
ATOM 2 C CA . MET A 1 1 ? 88.663 98.475 124.794 1 165.94 ? 1 A 1
"""


def test_parse_mmcif_filename_cif(sample_cif, tmp_path):
    fn = tmp_path / "sample.cif"
    fn.write_text(sample_cif)

    result = parse_mmcif(str(fn))

    assert len(result) == 19
    assert result["type_symbol"] == ["N", "C"]


def test_parse_mmcif_fileobj_cif(sample_cif, tmp_path):
    fn = tmp_path / "sample.cif"
    fn.write_text(sample_cif)

    with open(fn, "rb") as f:
        result = parse_mmcif(f)

    assert len(result) == 19
    assert result["type_symbol"] == ["N", "C"]


def test_parse_mmcif_filename_cif_gz(sample_cif, tmp_path):
    fn = tmp_path / "sample.cif.gz"
    with gzip.open(fn, "wt", encoding="utf-8") as f:
        f.write(sample_cif)

    result = parse_mmcif(str(fn))

    assert len(result) == 19
    assert result["type_symbol"] == ["N", "C"]


def test_parse_mmcif_fileobj_cif_gz(sample_cif, tmp_path):
    fn = tmp_path / "sample.cif.gz"
    with gzip.open(fn, "wt", encoding="utf-8") as f:
        f.write(sample_cif)

    with open(fn, "rb") as f:
        result = parse_mmcif(f)

    assert len(result) == 19
    assert result["type_symbol"] == ["N", "C"]


@pytest.fixture
def sample_pdb():
    return """\
ATOM      1  N   MET A   1      88.638  99.772 125.454  1.00165.94           N  
ATOM      2  CA  MET A   1      88.663  98.475 124.794  1.00165.94           C  
"""


def test_parse_pdb_filename_pdb(sample_pdb, tmp_path):
    fn = tmp_path / "sample.pdb"
    fn.write_text(sample_pdb)

    result = parse_pdb(str(fn))

    assert len(result) == 16
    assert result["name"] == ["N", "CA"]


def test_parse_pdb_fileobj_pdb(sample_pdb, tmp_path):
    fn = tmp_path / "sample.pdb"
    fn.write_text(sample_pdb)

    with open(fn) as f:
        result = parse_pdb(f)

    assert len(result) == 16
    assert result["name"] == ["N", "CA"]


def test_parse_pdb_filename_pdb_gz(sample_pdb, tmp_path):
    fn = tmp_path / "sample.pdb.gz"
    with gzip.open(fn, "wt", encoding="utf-8") as f:
        f.write(sample_pdb)

    result = parse_pdb(str(fn))

    assert len(result) == 16
    assert result["name"] == ["N", "CA"]


def test_parse_pdb_fileobj_pdb_gz(sample_pdb, tmp_path):
    fn = tmp_path / "sample.pdb.gz"
    with gzip.open(fn, "wt", encoding="utf-8") as f:
        f.write(sample_pdb)

    with open(fn, "rb") as f:
        result = parse_pdb(f)

    assert len(result) == 16
    assert result["name"] == ["N", "CA"]



import importlib.resources
from string import Template

import pyopencl as cl
from pyopencl.elementwise import ElementwiseKernel

import powerfit_em.correlators


class CLKernels:
    def __init__(self, ctx, values):
        self.sampler_nearest = cl.Sampler(ctx, True, cl.addressing_mode.REPEAT, cl.filter_mode.NEAREST)
        self.sampler_linear = cl.Sampler(ctx, True, cl.addressing_mode.REPEAT, cl.filter_mode.LINEAR)
        self.multiply = ElementwiseKernel(ctx, "float *x, float *y, float *z", "z[i] = x[i] * y[i];")
        self.conj_multiply = ElementwiseKernel(
            ctx, "cfloat_t *x, cfloat_t *y, cfloat_t *z", "z[i] = cfloat_mul(cfloat_conj(x[i]), y[i]);"
        )
        self.calc_lcc_and_take_best = ElementwiseKernel(
            ctx,
            """float *gcc, float *ave, float *ave2, int *mask,
                    float norm_factor, int nrot, float *lcc, int *grot""",
            """float _lcc;
                    if (mask[i] > 0) {
                        _lcc = gcc[i] / sqrt(ave2[i] * norm_factor - ave[i] * ave[i]);
                        if (_lcc > lcc[i]) {
                            lcc[i] = _lcc;
                            grot[i] = nrot;
                        };
                    };
                """,
        )

        t = importlib.resources.read_text(powerfit_em.correlators, "kernels.cl")
        t = Template(t).substitute(**values)

        self._program = cl.Program(ctx, t).build()
        self._rotate_image3d = self._program.rotate_image3d
        self._gws_rotate_grid3d = (96, 64, 1)

    def rotate_image3d(self, queue: cl.CommandQueue, image, rotmat, out, nearest=False):
        if nearest:
            args = (image, self.sampler_nearest, rotmat, out.data)
        else:
            args = (image, self.sampler_linear, rotmat, out.data)
        self._rotate_image3d(queue, self._gws_rotate_grid3d, None, *args)



import cython
import numpy as np
cimport numpy as np
from libc.math cimport ceil, exp, floor, sqrt, round


@cython.boundscheck(False)
def fsc_curve(np.ndarray[np.complex128_t, ndim=3] f1,
              np.ndarray[np.complex128_t, ndim=3] f2):

    cdef:
        int z, y, x, r, rmax
        double zmax, ymax, xmax
        np.ndarray[np.float64_t] f_1_2, f_1_1, f_2_2

    rmax = <int> np.max([f1.shape[n] for n in range(3)]) // 2
    f_1_2 = <np.ndarray[np.float64_t]> np.zeros(rmax)
    f_1_1 = <np.ndarray[np.float64_t]> np.zeros(rmax)
    f_2_2 = <np.ndarray[np.float64_t]> np.zeros(rmax)

    zmax = f1.shape[0] / 2.0
    ymax = f1.shape[1] / 2.0
    xmax = f1.shape[2] / 2.0

    for z in range(f1.shape[0]):
        if z > zmax:
            z -= f1.shape[0]
        for y in range(f1.shape[1]):
            if y > ymax:
                y -= f1.shape[1]
            for x in range(f1.shape[2]):
                if x > xmax:
                    x -= f1.shape[2]
                r = <int> round(sqrt(x * x + y * y + z * z))
                if r >= rmax:
                    continue

                f_1_2[r] += (f1[z, y, x] * f2[z, y, x].conjugate()).real
                f_1_1[r] += (f1[z, y, x] * f1[z, y, x].conjugate()).real
                f_2_2[r] += (f2[z, y, x] * f2[z, y, x].conjugate()).real

    fsc = f_1_2 / np.sqrt(f_1_1 * f_2_2)
    return fsc


@cython.boundscheck(False)
@cython.cdivision(True)
def blur_points(np.ndarray[np.float64_t, ndim=2] points,
                np.ndarray[np.float64_t, ndim=1] weights,
                double sigma,
                np.ndarray[np.float64_t, ndim=3] out,
                wraparound=False,
                ):
    """Convolutes points with a Gaussian kernel

    Parameters
    ----------

    Returns
    -------
    None

    """

    cdef:
        unsigned int n
        int x, y, z, xmin, ymin, zmin, xmax, ymax, zmax
        double extend, extend2, dsigma2
        double z2, y2z2, x2y2z2

    extend = 4.0 * sigma
    extend2 = extend * extend
    dsigma2 = 2.0 * sigma * sigma

    if wraparound:
        xmin_limit = -out.shape[2] + 1
        ymin_limit = -out.shape[1] + 1
        zmin_limit = -out.shape[0] + 1
    else:
        xmin_limit = 0
        ymin_limit = 0
        zmin_limit = 0

    xmax_limit = out.shape[2] - 1
    ymax_limit = out.shape[1] - 1
    zmax_limit = out.shape[0] - 1

    for n in range(points.shape[1]):

        xmin = <int> ceil(points[0, n] - extend)
        ymin = <int> ceil(points[1, n] - extend)
        zmin = <int> ceil(points[2, n] - extend)
        xmin = max(xmin, xmin_limit)
        ymin = max(ymin, ymin_limit)
        zmin = max(zmin, zmin_limit)

        xmax = <int> floor(points[0, n] + extend)
        ymax = <int> floor(points[1, n] + extend)
        zmax = <int> floor(points[2, n] + extend)
        xmax = min(xmax, xmax_limit)
        ymax = min(ymax, ymax_limit)
        zmax = min(zmax, zmax_limit)

        for z in range(zmin, zmax+1):
            z2 = (z - points[2, n])**2
            for y in range(ymin, ymax+1):
                y2z2 = (y - points[1, n])**2 + z2
                for x in range(xmin, xmax+1):
                    x2y2z2 = (x - points[0, n])**2 + y2z2
                    if x2y2z2 <= extend2:
                        out[z, y, x] += weights[n] * exp(-x2y2z2 / dsigma2)


@cython.boundscheck(False)
@cython.cdivision(True)
def dilate_points(
    np.ndarray[np.float64_t, ndim=2] points,
    np.ndarray[np.float64_t] radii,
    np.ndarray[np.float64_t, ndim=3] out,
    wraparound=False,
):
    """Dilate points into balls on a grid

    Parameters
    ----------

    Returns
    -------
    None

    """

    cdef:
        unsigned int n
        int x, y, z, xmin, ymin, zmin, xmax, ymax, zmax
        double radius, radius2
        double z2, y2z2, x2y2z2

    if wraparound:
        xmin_limit = -out.shape[2] + 1
        ymin_limit = -out.shape[1] + 1
        zmin_limit = -out.shape[0] + 1
    else:
        xmin_limit = 0
        ymin_limit = 0
        zmin_limit = 0

    xmax_limit = out.shape[2] - 1
    ymax_limit = out.shape[1] - 1
    zmax_limit = out.shape[0] - 1

    for n in range(points.shape[1]):

        radius = radii[n]
        radius2 = radius ** 2

        xmin = <int> ceil(points[0, n] - radius)
        ymin = <int> ceil(points[1, n] - radius)
        zmin = <int> ceil(points[2, n] - radius)
        xmin = max(xmin, xmin_limit)
        ymin = max(ymin, ymin_limit)
        zmin = max(zmin, zmin_limit)

        xmax = <int> floor(points[0, n] + radius)
        ymax = <int> floor(points[1, n] + radius)
        zmax = <int> floor(points[2, n] + radius)
        xmax = min(xmax, xmax_limit) + 1
        ymax = min(ymax, ymax_limit) + 1
        zmax = min(zmax, zmax_limit) + 1

        for z in range(zmin, zmax):
            z2 = (z - points[2, n])**2
            for y in range(ymin, ymax):
                y2z2 = (y - points[1, n])**2 + z2
                for x in range(xmin, xmax):
                    x2y2z2 = (x - points[0, n])**2 + y2z2
                    if x2y2z2 <= radius2:
                        out[z, y, x] = 1



import multiprocessing
from functools import partial
from multiprocessing import Lock, Process, RawValue
from multiprocessing.managers import DictProxy
from typing import TYPE_CHECKING, Any

import numpy as np
from tqdm.auto import tqdm

from powerfit_em.correlators.cpu import CPUCorrelator
from powerfit_em.helpers import opencl_available
from powerfit_em.volume import Volume

if TYPE_CHECKING:
    import pyopencl as cl  # noqa: I001


class _Counter:
    """Thread-safe counter object to follow PowerFit progress"""

    def __init__(self):
        self.val = RawValue("i", 0)
        self.lock = Lock()

    def increment(self):
        with self.lock:
            self.val.value += 1

    def value(self):
        with self.lock:
            return self.val.value


def run_correlator_instance(
    target: Volume,
    template: Volume,
    mask: Volume,
    rotations: np.ndarray,
    laplace: bool,
    jobid: int,
    counter: _Counter | None,
    results: dict[int, Any],
):
    correlator = CPUCorrelator(
        target.array,
        template.array,
        rotations,
        mask.array,
        laplace,
    )
    for n in range(len(rotations)):
        correlator.compute_rotation(n, rotmat=correlator.rotations[n])
        if counter is not None:
            counter.increment()

    results[jobid] = (correlator.lcc, correlator.rot)


class PowerFitter:
    """Wrapper around the Correlator classes for multiprocessing and GPU
    accelerated searches providing an easy interface.
    """

    def __init__(
        self,
        target: Volume,
        rotations: np.ndarray,
        template: Volume,
        mask: Volume,
        queue: "cl.CommandQueue | None",
        nproc: int = 1,
        laplace: bool = False,
    ):
        self._target = target
        self._rotations = rotations
        self._template = template
        self._mask = mask
        self._queue = queue
        self._nproc = nproc
        self._laplace = laplace
        self._corr = None
        self._lcc = np.zeros(0, dtype=np.float32)
        self._rot = np.zeros(0, dtype=np.float32)

    @property
    def lcc(self):
        return self._lcc.copy()

    @property
    def rot(self):
        return self._rot.copy()

    def scan(self, progress: partial[tqdm] | None):
        if self._queue is None:
            if self._nproc == 1:
                self._single_cpu_scan(progress)
            else:
                self._multi_cpu_scan(progress)
        else:
            self._gpu_scan(progress)

    def set_template(self, template: Volume, mask: Volume):
        if not self._corr:
            msg = "No correlator available yet. First run scan."
            raise ValueError(msg)
        self._corr.set_template(template.array, mask.array)

    def _gpu_scan(self, progress: partial[tqdm] | None):
        if opencl_available():
            from powerfit_em.correlators.gpu import GPUCorrelator

            if self._corr is None:
                self._corr = GPUCorrelator(
                    self._target.array,
                    self._template.array,
                    self._rotations,
                    self._mask.array,
                    self._queue,
                    self._laplace,
                )
            self._corr.scan(progress)
            self._lcc = self._corr.lcc
            self._rot = self._corr.rot
        else:
            raise ValueError("No OpenCL available")

    def _multi_cpu_scan(self, progress: partial[tqdm] | None):
        nrot = self._rotations.shape[0]
        self._nrot_per_job = nrot // self._nproc
        processes: list[Process] = []
        self._counter = None if processes is None else _Counter()
        self._njobs = self._nproc
        if self._queue is not None:
            self._njobs = len(self._queue)

        ids = tuple(range(self._njobs))
        manager = multiprocessing.Manager()
        results = manager.dict()

        for id in ids:
            start = id * self._nrot_per_job
            stop = start + self._nrot_per_job
            if id == self._njobs - 1:
                stop = len(self._rotations)
            partial_rotations = self._rotations[start:stop]
            processes.append(
                Process(
                    target=run_correlator_instance,
                    args=(
                        self._target,
                        self._template,
                        self._mask,
                        partial_rotations,
                        self._laplace,
                        id,
                        self._counter,
                        results,
                    ),
                )
            )

        for id in range(self._njobs):
            processes[id].start()

        if progress is not None:
            with progress(total=nrot) as pbar:
                while self._counter.value() < nrot:
                    current_count = self._counter.value()
                    pbar.update(current_count - pbar.n)

        for id in range(self._njobs):
            processes[id].join()
        self._combine(ids, results)

    def _single_cpu_scan(self, progress: partial[tqdm] | None):
        self._corr = CPUCorrelator(
            self._target.array,
            self._template.array,
            self._rotations,
            self._mask.array,
            self._laplace,
        )
        self._corr.scan(progress)
        self._lcc = self._corr.lcc
        self._rot = self._corr.rot

    def _combine(self, ids: tuple[int, ...], results: DictProxy):
        # Combine all the intermediate results
        lcc = np.zeros(self._target.shape)
        rot = np.zeros(self._target.shape)
        ind = np.zeros(lcc.shape, dtype=np.bool)
        for n in ids:
            # Get LCC and rotations from results
            part_lcc = results[n][0]
            part_rot = results[n][1]

            np.greater(part_lcc, lcc, ind)
            lcc[ind] = part_lcc[ind]
            # take care of the rotation index offset for each independent job
            rot[ind] = part_rot[ind] + self._nrot_per_job * n
        self._lcc = lcc
        self._rot = rot



from numpy import bool, greater_equal, log, zeros
from scipy.ndimage import label, maximum_position


class Analyzer:
    def __init__(
        self,
        corr,
        rotmat,
        rotmat_ind,
        steps=5,
        voxelspacing: float = 1.0,
        origin=(0, 0, 0),
        z_sigma: float = 1.0,
    ):
        self._corr = corr
        self._rotmat = rotmat
        self._rotmat_ind = rotmat_ind
        self._voxelspacing = voxelspacing
        self._origin = origin
        self._z_sigma = z_sigma
        self.steps = steps
        self._solutions = None

    @property
    def corr(self):
        return self._corr

    @property
    def rot(self):
        return self._rot

    @property
    def steps(self):
        return self._steps

    @steps.setter
    def steps(self, steps):
        self._steps = steps
        self._watershed()
        self._solutions = None

    @property
    def voxelspacing(self):
        return self._voxelspacing

    @voxelspacing.setter
    def voxelspacing(self, voxelspacing):
        self._solutions = None
        self._voxelspacing = voxelspacing

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, origin):
        self._solutions = None
        self._origin = origin

    @property
    def solutions(self) -> list[list[float]]:
        if self._solutions is None:
            self._generate_solutions()
        return self._solutions

    def _generate_solutions(self):
        solutions: list[list[float]] = []
        for pos in self._positions:
            solution: list[float] = []
            # To support scipy 0.14.0 cast pos into tuple of ints since
            # maximum_position returns a list of tuples of floats
            pos = tuple(int(p) for p in pos)
            lcc = self._corr[pos]
            solution.append(lcc)
            fishers_z = 0.5 * (log(1 + lcc) - log(1 - lcc))
            solution.append(fishers_z)
            rel_z = fishers_z / self._z_sigma
            solution.append(rel_z)
            z, y, x = [coor * self._voxelspacing + shift for coor, shift in zip(pos, self._origin[::-1], strict=False)]
            rotmat = self._rotmat[int(self._rotmat_ind[pos])]
            solution += [x, y, z] + list(rotmat.ravel())
            solutions.append(solution)
        self._solutions = sorted(solutions, key=lambda cc: cc[0], reverse=True)

    def _watershed(self):
        """Get positions of high correlation values using watershed
        algorithm
        """
        max_cc = self._corr.max()
        min_cc = 0.5 * max_cc
        stepsize = (max_cc - min_cc) / self._steps
        cutoff = max_cc
        positions = []
        mask = zeros(self._corr.shape, dtype=bool)
        for _ in range(self._steps):
            cutoff -= stepsize
            greater_equal(self._corr, cutoff, mask)
            labels, nfeatures = label(mask)
            positions += maximum_position(self._corr, labels, list(range(1, nfeatures + 1)))
        self._positions = set(positions)

    def tofile(self, out="solutions.out", delimiter=None):
        """Write solutions to file.

        Arguments:
            out: str, output file name (default: 'solutions.out')
            delimiter: str, delimiter for the output file. Defaults to fixed width output.
                Can be ',' or '\t'. With delimiter set the header will not have leading '#'.
        """

        if self._solutions is None:
            self._generate_solutions()

        write_file(self._solutions, out, delimiter)


def write_file(solutions: list[list[str | float]], out: str, delimiter: str | None):
    fit_many = len(solutions[0]) == 16
    headers = [
        "#rank",
        "cc",
        "Fish-z",
        "rel-z",
        "x",
        "y",
        "z",
        "a11",
        "a12",
        "a13",
        "a21",
        "a22",
        "a23",
        "a31",
        "a32",
        "a33",
    ]

    max_name_length = 0
    if fit_many:
        headers.insert(1, "pdb-filename")
        max_name_length = max([len(x[0]) for x in solutions])

    with open(out, "w") as f:
        if delimiter is None:
            header_fmt = ["{:<6s}"] + ["{:>6s}"] * 3 + ["{:>8s}"] * 3 + ["{:>6s}"] * 9
            if fit_many:
                header_fmt.insert(1, "{:<" + str(max_name_length) + "s}")
            line = " ".join(header_fmt) + "\n"
            f.write(line.format(*headers))

            line_fmt = ["{:<6d}"] + ["{:6.3f}"] * 3 + ["{:8.3f}"] * 3 + ["{:6.3f}"] * 9
            if fit_many:
                line_fmt.insert(1, "{:<" + str(max_name_length) + "s}")
            line = " ".join(line_fmt) + "\n"
            for n, sol in enumerate(solutions):
                f.write(line.format(n + 1, *sol))
        else:
            # Write header
            headers[0] = headers[0].lstrip("#")
            f.write(delimiter.join(headers) + "\n")
            for n, sol in enumerate(solutions):
                row = [n + 1] + sol
                # Format all values as floats with 3 decimals except rank (int)
                row_str = (
                    [str(row[0])]
                    + [f"{v:.3f}" for v in row[1:4]]
                    + [f"{v:.3f}" for v in row[4:7]]
                    + [f"{v:.3f}" for v in row[7:]]
                )
                if fit_many:
                    row_str.insert(1, str(row[1]))
                f.write(delimiter.join(row_str) + "\n")



from importlib.util import find_spec
from math import sqrt

import numpy as np
from scipy.ndimage import binary_erosion


def pyfftw_available() -> bool:
    return find_spec("pyfftw") is not None


def opencl_available() -> bool:
    return find_spec("pyopencl") is not None


def determine_core_indices(mask):
    """Calculate the core indices of a shape"""

    core_indices = np.zeros(mask.shape)
    eroded_mask = mask > 0
    while eroded_mask.sum() > 0:
        core_indices += eroded_mask
        eroded_mask = binary_erosion(eroded_mask)
    return core_indices


def fisher_sigma(mv, fsc):
    return 1 / sqrt(mv / fsc - 3)


def write_fits_to_pdb(structure, solutions, basename="fit"):
    translated_structure = structure.duplicate()
    center = translated_structure.coor.mean(axis=1)
    translated_structure.translate(-center)
    for n, sol in enumerate(solutions, start=1):
        out = translated_structure.duplicate()
        rot = np.asarray([float(x) for x in sol[6:]]).reshape(3, 3)
        trans = sol[3:6]
        out.rotate(rot)
        out.translate(trans)
        out.tofile(basename + f"_{n:d}.pdb")
